from typing import Dict, Tuple

from ..types import Scope
from . import EmptyInformation


async def from_session(scope: Scope) -> Tuple[str, str]:
    """
    get user id and group from session (need starlette SessionMiddleware)
    """
    assert (
        "session" in scope
    ), "Starlette SessionMiddleware must be installed to access request.session"
    session: Dict[str, str] = scope["session"]
    try:
        return session["user"], session.get("group", "default")
    except KeyError:
        raise EmptyInformation(scope)



import json
import time

from redis.asyncio import StrictRedis

from ..rule import Rule
from . import BaseBackend

SLIDING_WINDOW_SCRIPT = """
-- Set variables from arguments
local now = tonumber(ARGV[1])
local ruleset = cjson.decode(ARGV[2])
local result = {}
-- ruleset looks like this:
-- {key: [limit, window_size], ...}
local scores = {}
local min = {}
local expire_in = {}
for i, pgname in ipairs(KEYS) do
    -- we remove keys older than now - window_size
    local clearBefore = now - ruleset[pgname][2]
    redis.call('ZREMRANGEBYSCORE', pgname, 0, clearBefore)
    -- we get the count
    local amount = redis.call('ZCARD', pgname)
    -- we add to sorted set if allowed ie the amount < limit
    if amount < ruleset[pgname][1] then
        redis.call('ZADD', pgname, now, now)
    end
    -- cleanup, this expires the whole set in window_size secs
    redis.call('EXPIRE', pgname, ruleset[pgname][2])
    -- calculate the remaining amount of requests. If >= 0 then request for that window is allowed
    scores[i] = ruleset[pgname][1] - amount
    min[i] = redis.call('ZRANGE', pgname, 0, 0)[1]
    expire_in[i] = - now + tonumber(min[i]) + ruleset[pgname][2]
end
result['scores'] = scores
result['expire_in'] = expire_in
return cjson.encode(result)
"""


class SlidingRedisBackend(BaseBackend):
    def __init__(
        self,
        host: str = "localhost",
        port: int = 6379,
        db: int = 0,
        password: str = None,
        ssl: bool = False,
    ) -> None:
        self._redis = StrictRedis(
            host=host, port=port, db=db, password=password, ssl=ssl
        )
        self.sliding_function = self._redis.register_script(SLIDING_WINDOW_SCRIPT)

    async def get_limits(self, path: str, user: str, rule: Rule) -> dict:
        epoch = time.time()
        ruleset = rule.ruleset(path, user)
        r = await self.sliding_function(
            keys=list(ruleset.keys()), args=[epoch, json.dumps(ruleset)]
        )
        mr = json.loads(r.decode())
        # we need that in case redis returns no values for a given key, "scores" or "expire_in"
        # if that is the case the corresponding value will be {} and we transform it to []
        mr = {k: (v if v != {} else []) for k, v in mr.items()}
        mr["epoch"] = epoch
        return mr

    async def set_block_time(self, user: str, block_time: int) -> None:
        await self._redis.set(f"blocking:{user}", 1, block_time)

    async def is_blocking(self, user: str) -> int:
        return int(await self._redis.ttl(f"blocking:{user}"))

    async def retry_after(self, path: str, user: str, rule: Rule) -> int:
        block_time = await self.is_blocking(user)
        if block_time > 0:
            return block_time

        limits = await self.get_limits(path, user, rule)
        retry_after = limits["expire_in"][0] if not all(limits["scores"]) else 0

        if retry_after > 0 and rule.block_time:
            await self.set_block_time(user, rule.block_time)
            retry_after = rule.block_time

        return round(retry_after)



import json

from redis.asyncio import StrictRedis
from redis.commands.core import Script

from ..rule import Rule
from . import BaseBackend

SCRIPT = """
local ruleset = cjson.decode(ARGV[1])

-- Set limits
for i, key in pairs(KEYS) do
    redis.call('SET', key, ruleset[key][1], 'EX', ruleset[key][2], 'NX')
end

-- Check limits
for i = 1, #KEYS do
    local value = redis.call('GET', KEYS[i])
    if value and tonumber(value) < 1 then
        return ruleset[KEYS[i]][2]
    end
end

-- Decrease limits
for i, key in pairs(KEYS) do
    redis.call('DECR', key)
end
return 0
"""


class RedisBackend(BaseBackend):
    def __init__(
        self,
        host: str = "localhost",
        port: int = 6379,
        db: int = 0,
        password: str = None,
        ssl: bool = False,
    ) -> None:
        self._redis = StrictRedis(
            host=host, port=port, db=db, password=password, ssl=ssl
        )
        self.lua_script: Script = self._redis.register_script(SCRIPT)

    async def set_block_time(self, user: str, block_time: int) -> None:
        await self._redis.set(f"blocking:{user}", 1, block_time)

    async def is_blocking(self, user: str) -> int:
        return int(await self._redis.ttl(f"blocking:{user}"))

    async def retry_after(self, path: str, user: str, rule: Rule) -> int:
        block_time = await self.is_blocking(user)
        if block_time > 0:
            return block_time

        ruleset = rule.ruleset(path, user)
        retry_after = int(
            await self.lua_script(keys=list(ruleset.keys()), args=[json.dumps(ruleset)])
        )

        if retry_after > 0 and rule.block_time:
            await self.set_block_time(user, rule.block_time)
            retry_after = rule.block_time

        return retry_after



from typing import Awaitable, Callable, List, Tuple, TypeVar, Union

import jwt

try:
    from cryptography.hazmat.primitives.asymmetric.rsa import (
        RSAPrivateKey,
        RSAPublicKey,
    )
except ImportError:  # pragma: no cover
    RSAPublicKey = TypeVar("RSAPublicKey")
    RSAPrivateKey = TypeVar("RSAPrivateKey")

from ..types import Scope
from . import EmptyInformation


def create_jwt_auth(
    key: Union[bytes, str, RSAPublicKey, RSAPrivateKey],
    algorithms: Union[List[str], str],
    user_key: str = "user",
    group_key: str = "group",
) -> Callable[[Scope], Awaitable[Tuple[str, str]]]:
    """
    create jwt authentication function

    * key: for algorithm secret key
    * algorithms: Possible algorithms in https://pyjwt.readthedocs.io/en/latest/algorithms.html#digital-signature-algorithms
    """

    async def jwt_auth(scope: Scope) -> Tuple[str, str]:
        """
        About jwt header, read this link:
        https://stackoverflow.com/questions/33265812/best-http-authorization-header-type-for-jwt
        """
        for name, value in scope["headers"]:  # type: bytes, bytes
            if name == b"authorization":
                authorization = value.decode("utf8")
                break
        else:
            authorization = None

        if not authorization:
            raise EmptyInformation(scope)

        token_type, json_web_token = authorization.split(" ")

        assert (
            token_type == "Bearer"
        ), "Authorization header must be `Bearer` type. Like: `Bearer LONG_JWT`"

        data = jwt.decode(json_web_token, key, algorithms=algorithms)

        try:
            return data[user_key], data.get(group_key, "default")
        except KeyError:
            raise EmptyInformation(scope)

    return jwt_auth



from ipaddress import ip_address
from typing import Tuple

from ..types import Scope
from . import EmptyInformation


async def client_ip(scope: Scope) -> Tuple[str, str]:
    """
    parse ip
    """
    real_ip = ""
    if scope["client"]:
        ip, port = tuple(scope["client"])
        if ip_address(ip).is_global:
            real_ip = ip
    else:
        raise EmptyInformation(scope)

    for name, value in scope["headers"]:  # type: bytes, bytes
        if name == b"x-real-ip":
            ip = value.decode("utf8")

        if not real_ip and ip_address(ip).is_global:
            real_ip = ip

    if not real_ip:
        raise EmptyInformation(scope)
    return real_ip, "default"



from abc import ABC, abstractmethod

from ..rule import Rule


class BaseBackend(ABC):
    """
    Base class for all backend
    """

    @abstractmethod
    async def retry_after(self, path: str, user: str, rule: Rule) -> int:
        raise NotImplementedError



import asyncio
import re
from typing import Awaitable, Callable, Dict, Optional, Sequence, Tuple

from .backends import BaseBackend
from .rule import RULENAMES, Rule
from .types import ASGIApp, Receive, Scope, Send


def _on_blocked(retry_after: int) -> ASGIApp:
    async def default_429(scope: Scope, receive: Receive, send: Send) -> None:
        await send(
            {
                "type": "http.response.start",
                "status": 429,
                "headers": [
                    (b"retry-after", str(retry_after).encode("ascii")),
                ],
            }
        )
        await send({"type": "http.response.body", "body": b"", "more_body": False})

    return default_429


class RateLimitMiddleware:
    """
    rate limit middleware
    """

    def __init__(
        self,
        app: ASGIApp,
        authenticate: Callable[[Scope], Awaitable[Tuple[str, str]]],
        backend: BaseBackend,
        config: Dict[str, Sequence[Rule]],
        *,
        on_auth_error: Optional[Callable[[Exception], Awaitable[ASGIApp]]] = None,
        on_blocked: Callable[[int], ASGIApp] = _on_blocked,
    ) -> None:

        self.app = app
        self.authenticate = authenticate
        self.backend = backend

        if not asyncio.iscoroutinefunction(self.authenticate):
            raise ValueError(f"invalid authenticate function: {self.authenticate}")

        assert isinstance(backend, BaseBackend), f"invalid backend: {self.backend}"

        self.config: Dict[re.Pattern, Sequence[Rule]] = {
            re.compile(path): value for path, value in config.items()
        }

        self.on_auth_error = on_auth_error
        self.on_blocked = on_blocked

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":  # pragma: no cover
            return await self.app(scope, receive, send)

        url_path = scope["path"]
        for pattern, rules in self.config.items():
            if not pattern.match(url_path):
                continue
            # After finding the first rule that can match the path,
            # calculate the user ID and group
            try:
                user, group = await self.authenticate(scope)
            except Exception as exc:
                if self.on_auth_error is not None:
                    response = await self.on_auth_error(exc)
                    return await response(scope, receive, send)
                raise exc

            # Select the first rule that can be matched
            match_rule = list(filter(lambda r: r.group == group, rules))
            if match_rule:
                rule = match_rule[0]
                break
        else:  # If no rule can match, run `self.app` and return
            return await self.app(scope, receive, send)

        if not any(getattr(rule, name) is not None for name in RULENAMES):
            return await self.app(scope, receive, send)

        path: str = url_path if rule.zone is None else rule.zone
        retry_after = await self.backend.retry_after(path, user, rule)
        if retry_after == 0:
            return await self.app(scope, receive, send)

        return await self.on_blocked(retry_after)(scope, receive, send)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ratelimit', 'ratelimit.auths', 'ratelimit.backends']

package_data = \
{'': ['*']}

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.6,<0.7'],
 'full': ['pyjwt>=1.7.1,<2.0.0', 'redis>=4.2.0rc2'],
 'jwt': ['pyjwt>=1.7.1,<2.0.0'],
 'redis': ['redis>=4.2.0rc2']}

setup_kwargs = {
    'name': 'asgi-ratelimit',
    'version': '0.8.0',
    'description': '',
    'long_description': '# ASGI RateLimit\n\nLimit user access frequency. Base on ASGI.\n\n100% coverage. High performance. Support regular matching. Customizable.\n\n## Install\n\n```\n# Only install\npip install asgi-ratelimit\n\n# Use redis\npip install asgi-ratelimit[redis]\n\n# Use jwt\npip install asgi-ratelimit[jwt]\n\n# Install all\npip install asgi-ratelimit[full]\n```\n\n## Usage\n\nThe following example will limit users under the `"default"` group to access `/towns` at most once per second and `/forests` at most once per minute. And the users in the `"admin"` group have no restrictions.\n\n```python\nfrom typing import Tuple\n\nfrom ratelimit import RateLimitMiddleware, Rule\nfrom ratelimit.auths import EmptyInformation\nfrom ratelimit.backends.redis import MemoryBackend, RedisBackend\n\n# Simple rate-limiter in memory:\n\nrate_limit = RateLimitMiddleware(\n    ASGI_APP,\n    AUTH_FUNCTION,\n    MemoryBackend(),\n    {\n        r"^/towns": [Rule(second=1, group="default"), Rule(group="admin")],\n        r"^/forests": [Rule(minute=1, group="default"), Rule(group="admin")],\n    },\n)\n\n# with Redis:\n\nrate_limit = RateLimitMiddleware(\n    ASGI_APP,\n    AUTH_FUNCTION,\n    RedisBackend(),\n    {\n        r"^/towns": [Rule(second=1, group="default"), Rule(group="admin")],\n        r"^/forests": [Rule(minute=1, group="default"), Rule(group="admin")],\n    },\n)\n\n# Or if using Starlette, FastApi, or index.py framework\napp.add_middleware(\n    RateLimitMiddleware,\n    authenticate=AUTH_FUNCTION,\n    backend=RedisBackend(),\n    config={\n        r"^/towns": [Rule(second=1, group="default"), Rule(group="admin")],\n        r"^/forests": [Rule(minute=1, group="default"), Rule(group="admin")],\n    },\n)\n```\n\n:warning: **The pattern\'s order is important, rules are set on the first match**: Be careful here !\n\nNext, provide a custom authenticate function, or use one of the [existing auth methods](#built-in-auth-functions).\n\n```python\nasync def AUTH_FUNCTION(scope: Scope) -> Tuple[str, str]:\n    """\n    Resolve the user\'s unique identifier and the user\'s group from ASGI SCOPE.\n\n    If there is no user information, it should raise `EmptyInformation`.\n    If there is no group information, it should return "default".\n    """\n    # FIXME\n    # You must write the logic of this function yourself,\n    # or use the function in the following document directly.\n    return USER_UNIQUE_ID, GROUP_NAME\n\n\nrate_limit = RateLimitMiddleware(ASGI_APP, AUTH_FUNCTION, ...)\n```\n\nThe `Rule` type takes a time unit (e.g. `"second"`) and/or a `"group"`, as a param. If the `"group"` param is not specified then the `"authenticate"` method needs to return the "default group".\n\nExample:\n```python\n    ...\n    config={\n        r"^/towns": [Rule(second=1), Rule(second=10, group="admin")],\n    }\n    ...\n\n\nasync def AUTH_FUNCTION(scope: Scope) -> Tuple[str, str]:\n    ...\n    # no group information about this user\n    if user not in admins_group:\n        return user_unique_id, \'default\'\n\n    return user_unique_id, user_group\n```\n\n### Customizable rules\n\nIt is possible to mix the rules to obtain higher level of control.\n\nThe below example will allow up to 10 requests per second and no more than 200 requests per minute, for everyone, for the same API endpoint.\n\n```python\n    ...\n    config={\n        r"^/towns": [Rule(minute=200, second=10)],\n    }\n    ...\n```\n\nExample for a "admin" group with higher limits.\n\n```python\n    ...\n    config={\n        r"^/towns": [\n            Rule(day=400, minute=200, second=10),\n            Rule(minute=500, second=25, group="admin"),\n        ],\n    }\n    ...\n```\n\nSometimes you may want to specify that some APIs share the same flow control pool. In other words, flow control is performed on the entire set of APIs instead of a single specific API. Only the `zone` parameter needs to be used. **Note**: You can give different rules the same `zone` value, and all rules with the same `zone` value share the same flow control pool.\n\n```python\n    ...\n    config={\n        r"/user/\\d+": [\n            Rule(minute=200, zone="user-api"),\n            Rule(second=100, zone="user-api", group="admin"),\n        ],\n    }\n    ...\n```\n\n### Block time\n\nWhen the user\'s request frequency triggers the upper limit, all requests in the following period of time will be returned with a `429` status code.\n\nExample: `Rule(second=5, block_time=60)`, this rule will limit the user to a maximum of 5 visits per second. Once this limit is exceeded, all requests within the next 60 seconds will return `429`.\n\n### Custom block handler\n\nJust specify `on_blocked` and you can customize the asgi application that is called when blocked.\n\n```python\ndef yourself_429(retry_after: int):\n    async def inside_yourself_429(scope: Scope, receive: Receive, send: Send) -> None:\n        await send({"type": "http.response.start", "status": 429})\n        await send(\n            {\n                "type": "http.response.body",\n                "body": b"custom 429 page",\n                "more_body": False,\n            }\n        )\n\n    return inside_yourself_429\n\nRateLimitMiddleware(..., on_blocked=yourself_429)\n```\n\n### Built-in auth functions\n\n#### Client IP\n\n```python\nfrom ratelimit.auths.ip import client_ip\n```\n\nObtain user IP through `scope["client"]` or `X-Real-IP`.\n\nNote: this auth method will not work if your IP address (such as 127.0.0.1 etc) is not allocated for public networks.\n\n#### Starlette Session\n\n```python\nfrom ratelimit.auths.session import from_session\n```\n\nGet `user` and `group` from `scope["session"]`.\n\nIf key `group` not in session, will return `default`. If key `user` not in session, will raise a `EmptyInformation`.\n\n#### Json Web Token\n\n```python\nfrom ratelimit.auths.jwt import create_jwt_auth\n\njwt_auth = create_jwt_auth("KEY", "HS256")\n```\n\nGet `user` and `group` from JWT that in `Authorization` header.\n\n### Custom auth error handler\n\nNormally exceptions raised in the authentication function result in an Internal Server Error, but you can pass a function to handle the errors and send the appropriate response back to the user. For example, if you\'re using FastAPI or Starlette:\n\n```python\nfrom fastapi.responses import JSONResponse\nfrom ratelimit.types import ASGIApp\n\nasync def handle_auth_error(exc: Exception) -> ASGIApp:\n    return JSONResponse({"message": "Unauthorized access."}, status_code=401)\n\nRateLimitMiddleware(..., on_auth_error=handle_auth_error)\n```\n\nFor advanced usage you can handle the response completely by yourself:\n\n```python\nfrom fastapi.responses import JSONResponse\nfrom ratelimit.types import ASGIApp, Scope, Receive, Send\n\nasync def handle_auth_error(exc: Exception) -> ASGIApp:\n    async def response(scope: Scope, receive: Receive, send: Send):\n        # do something here e.g.\n        # await send({"type": "http.response.start", "status": 429})\n    return response\n```\n',
    'author': 'abersheeran',
    'author_email': 'me@abersheeran.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/abersheeran/asgi-ratelimit',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import Dict, Optional, Tuple


@dataclass
class Rule:
    group: str = "default"

    second: Optional[int] = None
    minute: Optional[int] = None
    hour: Optional[int] = None
    day: Optional[int] = None
    month: Optional[int] = None

    block_time: Optional[int] = None

    zone: Optional[str] = None

    def ruleset(self, path: str, user: str) -> Dict[str, Tuple[int, int]]:
        """
        builds a dictionary of keys, values where keys are
        the redis keys and values is a tuple of (limit, ttl)
        """
        return {
            f"{path}:{user}:{name}": (limit, TTL[name])
            for name, limit in map(lambda name: (name, getattr(self, name)), RULENAMES)
            if limit is not None
        }


TTL = {
    "second": 1,
    "minute": 60,
    "hour": 60 * 60,
    "day": 24 * 60 * 60,
    "month": 31 * 24 * 60 * 60,
}

RULENAMES: Tuple[str, ...] = ("second", "minute", "hour", "day", "month")



import asyncio
from collections import defaultdict
from dataclasses import dataclass
from threading import Lock
from typing import Dict, List, Optional

from ..rule import Rule
from . import BaseBackend


@dataclass
class Limit:
    count: int
    timestamp: int

    def decr(self):
        self.count -= 1

    def is_invalid(self, now):
        return self.count < 1 and self.timestamp > now


class MemoryBackend(BaseBackend):
    """simple limiter with memory"""

    def __init__(self):
        # user: deadline
        self.blocked_users: Dict[str, int] = {}
        # path: {rule_key: (limit, timestamp)}
        self.blocks: Dict[str, Dict[str, Limit]] = defaultdict(dict)

        self.blocks_lock = Lock()
        self.blocked_users_lock = Lock()

    @staticmethod
    def now() -> int:
        loop = asyncio.get_event_loop()
        return int(loop.time())

    @staticmethod
    def call_later(later, callback, *args) -> asyncio.TimerHandle:
        loop = asyncio.get_event_loop()
        return loop.call_later(later, callback, *args)

    def is_blocking(self, user: str) -> int:
        end_ts: int = self.blocked_users.get(user, 0)
        return max(end_ts - self.now(), 0)

    def remove_user(self, user: str) -> Optional[int]:
        with self.blocked_users_lock:
            return self.blocked_users.pop(user, None)

    def remove_rule(self, path: str, rule_key: str) -> Optional[List[int]]:
        with self.blocks_lock:
            return self.blocks[path].pop(rule_key, None)

    def remove_blocked_user_later(self, user: str) -> None:
        later = self.blocked_users[user]
        self.call_later(later, self.remove_user, user)

    def remove_rule_later(self, path: str, rule_key: str) -> None:
        limit_dataclass = self.blocks[path][rule_key]
        self.call_later(limit_dataclass.timestamp, self.remove_rule, path, rule_key)

    def set_blocked_user(self, user: str, block_time: int) -> int:
        self.blocked_users[user] = block_time + self.now()
        self.remove_blocked_user_later(user)
        return block_time

    def set_rule(
        self,
        rules: Dict[str, List[int]],
        path: str,
        rule: str,
        limit: int,
        timestamp: int,
    ) -> None:
        rules[rule] = Limit(limit - 1, timestamp)
        self.remove_rule_later(path, rule)

    async def retry_after(self, path: str, user: str, rule: Rule) -> int:
        block_time = self.is_blocking(user)
        if block_time > 0:
            return block_time
        ruleset = rule.ruleset(path, user)

        now = self.now()
        rules = self.blocks.setdefault(path, {})

        retry_after: int = 0

        for rule_, (limit, seconds) in ruleset.items():
            exist_rule = rules.get(rule_)
            if not exist_rule:
                self.set_rule(rules, path, rule_, limit, now + seconds)
            else:
                if exist_rule.is_invalid(now):
                    retry_after = exist_rule.timestamp - now
                    break
                if exist_rule.timestamp > now:
                    exist_rule.decr()
                else:
                    self.set_rule(rules, path, rule_, limit, now + seconds)

        if retry_after > 0 and rule.block_time:
            retry_after = self.set_blocked_user(user, rule.block_time)

        return retry_after
